/*
	* Copyright (c) 2021-11 by ELECTRONIC ASSEMBLY <technik@lcd-module.de>
	*
	* utf8 to unicode and vice versa
	*
	* This file is free software; you can redistribute it and/or modify
	* it under the terms of either the GNU General Public License version 2
	* or the GNU Lesser General Public License version 2.1, both as
	* published by the Free Software Foundation.
*/

#include <stdio.h>
#include <stdlib.h>
#include <locale.h>
#include "Unicode.h"

#define __STD_UTF_16__

/*----------------------------
	Func: utf8_to_unicode
	Desc: changes utf8 code to 16 bit unicode
	Vars: uft 8 string with size, pointer to unicode buffer with maximum length
	Ret : length of unicode string
------------------------------*/
int utf8_to_unicode(unsigned char* const utf8_str, int utf8_size, char16_t* unicode_output, int unicode_size) 
{
	unsigned char* p_utf8_Code = utf8_str;
	char16_t* p_unicode_Code = unicode_output;
	int utf8_iter = 0;
	int unicode_iter = 0;
	
	while (*p_utf8_Code && (unicode_iter < unicode_size || utf8_iter < utf8_size)) 
	{
		if (*p_utf8_Code < 0x80) //ASCII
		{
			
			*p_unicode_Code++ = (char16_t) (*p_utf8_Code++);
			
			unicode_iter++;
			utf8_iter++;
		}
		else if (*p_utf8_Code < 0xC2) //Ignore
		{
			p_utf8_Code++;
			utf8_iter++;
		}
		else if (*p_utf8_Code < 0xE0) //ASCII extended
		{
			char16_t hChar = (char16_t) ((*p_utf8_Code++) & 0x1F);	
			char16_t lChar = (char16_t) ((*p_utf8_Code++) & 0x3F);		
			
			*p_unicode_Code++ = (char16_t) ((hChar << 6) | lChar);
			
			unicode_iter++;
			utf8_iter += 2;
		}
		else if (*p_utf8_Code < 0xF0) //Unicode 3 bytes
		{
			char16_t hChar = (char16_t) ((*p_utf8_Code++) & 0xF);
			char16_t mChar = (char16_t) ((*p_utf8_Code++) & 0x3F);
			char16_t lChar = (char16_t) ((*p_utf8_Code++) & 0x3F);
				
			//Directly set the value to the UTF-16 code unit.
			*p_unicode_Code++ = (char16_t) ((hChar << 12) | (mChar << 6) | (lChar));

			unicode_iter++;
			utf8_iter += 3;
		}
		else if (*p_utf8_Code < 0xF8)  //Unicode 4 bytes
		{
			//only applicable with unicode 1.0.1
			/*char16_t hhChar = (char16_t) ((*p_utf8_Code++) & 0x07);
			char16_t hlChar = (char16_t) ((*p_utf8_Code++) & 0x3F);
			char16_t lhChar = (char16_t) ((*p_utf8_Code++) & 0x3F);
			char16_t llChar = (char16_t) ((*p_utf8_Code++) & 0x3F);
			
			*p_unicode_Code++ = (char16_t) ((hhChar << 18) | (hlChar << 12) | (lhChar << 6) | (llChar));
			unicode_iter++;		*/
			
			p_utf8_Code += 4;
			utf8_iter += 4;
		}
		else //Ignore
		{
			p_utf8_Code++;
			utf8_iter++;
		}
	}
	
	if (unicode_iter < unicode_size) //add String end
	{
		*p_unicode_Code = '\0';
	}
	
	return unicode_iter; //return length
}

/*----------------------------
	Func: unicode_to_utf8
	Desc: changes unicode to utf8
	Vars: unicode string with size, pointer to utf8 buffer with maximum length
	Ret : length of utf8 string
------------------------------*/
int unicode_to_utf8(char16_t *unicode_str, int unicode_size, unsigned char* utf8_output, int utf8_size)
{
	char16_t * p_unicode_Code = unicode_str;
	unsigned char* p_utf8_Code = utf8_output;
	int unicode_iter = 0;
	int utf8_iter = 0;
	
	while (*p_unicode_Code && (utf8_iter < utf8_size || unicode_iter < unicode_size)) 
	{
		if (*p_unicode_Code < 0x80) //ASCII
		{
			*p_utf8_Code++ = (unsigned char) (*p_unicode_Code++);
			utf8_iter++;	
			unicode_iter++;
		}
		else if (*p_unicode_Code < 0x800) //ASCII extended
		{		
			unsigned char fChar = (unsigned char)(((*p_unicode_Code >> 6) & 0x1F) | 0xC0 );
			unsigned char sChar = (unsigned char)((*p_unicode_Code++ & 0x3F) | 0x80);
			
			*p_utf8_Code++ = (unsigned char) fChar;
			*p_utf8_Code++ = (unsigned char) sChar;
			
			utf8_iter+=2;
		}
		else if (*p_unicode_Code < 0xFFFF) //Unicode 3 Bytes
		{		
			unsigned char fChar = (unsigned char)(((*p_unicode_Code >> 12) & 0x0F) | 0xE0 );
			unsigned char sChar = (unsigned char)(((*p_unicode_Code >> 6) & 0x3F) | 0x80 );
			unsigned char tChar = (unsigned char)((*p_unicode_Code++ & 0x3F) | 0x80);
						
			*p_utf8_Code++ = (unsigned char) fChar;
			*p_utf8_Code++ = (unsigned char) sChar;
			*p_utf8_Code++ = (unsigned char) tChar;
			
			utf8_iter+=3;
		}
	}
	
	if (utf8_iter < utf8_size) //add String end
	{
		*p_utf8_Code = '\0';
	}
	
	return utf8_iter; //return length
}